/*
 * $Id: metriczECKD.c,v 1.2 2009/05/20 19:39:56 tyreld Exp $
 *
 * (C) Copyright IBM Corp. 2006, 2009
 *
 * THIS FILE IS PROVIDED UNDER THE TERMS OF THE ECLIPSE PUBLIC LICENSE
 * ("AGREEMENT"). ANY USE, REPRODUCTION OR DISTRIBUTION OF THIS FILE
 * CONSTITUTES RECIPIENTS ACCEPTANCE OF THE AGREEMENT.
 *
 * You can obtain a current copy of the Eclipse Public License from
 * http://www.opensource.org/licenses/eclipse-1.0.php
 *
 * Author:       Alexander Wolf-Reber <a.wolf-reber@de.ibm.com>
 * Contributors:
 *
 * Description:
 * Metrics Gatherer Plugin for System z ECKD volume metrics
 *
 */

/* ---------------------------------------------------------------------------*/

#include <mplugin.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <commutil.h>
#include <dirent.h>

#include "sysfswrapper.h"

#define BUS "ccw"
#define DRIVER "dasd-eckd"

#ifdef DEBUG
#define DBGONLY(X)   X
#else
#define DBGONLY(X) 
#endif


inline int max(int x, int y) {
	return (x>y) ? x : y;
}

inline int min(int x, int y) {
	return (x>y) ? x : y;
}

/* ---------------------------------------------------------------------------*/

static MetricDefinition metricDef[1];

static MetricRetriever  metricRetrECKD;

/* ---------------------------------------------------------------------------*/

int _DefinedMetrics( MetricRegisterId *mr,
		     const char * pluginname,
		     size_t *mdnum,
		     MetricDefinition **md ) {

  DBGONLY(fprintf(stderr,"--- %s(%i) : Retrieving metric definitions\n",
	  __FILE__,__LINE__);)

  if (mr==NULL||mdnum==NULL||md==NULL) {
    fprintf(stderr,"--- %s(%i) : invalid parameter list\n",__FILE__,__LINE__);
    return -1;
  }

  metricDef[0].mdVersion=MD_VERSION;
  metricDef[0].mdName="_ECKD";
  metricDef[0].mdReposPluginName="librepositoryzECKD.so";
  metricDef[0].mdId=mr(pluginname,metricDef[0].mdName);
  metricDef[0].mdSampleInterval=60;
  metricDef[0].mproc=metricRetrECKD;
  metricDef[0].mdeal=free;

  *mdnum=1;
  *md=metricDef;
  return 0;
}

int _StartStopMetrics (int starting) {

  if( starting ) {
	  DBGONLY(fprintf(stderr,"--- %s(%i) : Metric processing starting\n",
			  __FILE__,__LINE__);)
    /* nothing to do */
  }
  else {
	  DBGONLY(fprintf(stderr,"--- %s(%i) : Metric processing stopping\n",
			  __FILE__,__LINE__);)
    /* nothing to do */
  }

  return 0;
}

int sendMetric(int mid, MetricReturner mret, char* device, char* values) {
	  MetricValue * mv  = NULL;

	  if (mret==NULL) {
		  fprintf(stderr,"Returner pointer is NULL\n");
		  return -1;
	  }

	  mv = calloc(1, sizeof(MetricValue) +
		      strlen(values) + 1 +
		      strlen(device) + 1 );
	  if (mv) {
		  mv->mvId = mid;
		  mv->mvTimeStamp = time(NULL);
		  mv->mvDataType = MD_STRING;
		  mv->mvDataLength = (strlen(values)+1);
		  mv->mvData = (char*)mv + sizeof(MetricValue);
		  strncpy( mv->mvData, values, strlen(values) );
		  mv->mvResource = (char*)mv + sizeof(MetricValue) +
			  strlen(values) + 1;
		  strcpy(mv->mvResource, device);
		  mret(mv);
		  return 1;
	  }
	  return -1;
}

char* readSysFs(char* deviceName) {
	int i, attributeCount;
	size_t strLength;
	const char* CMB_ENABLE = "cmb_enable";
	const char* SAMPLE_COUNT = "cmf/sample_count";
	const char* ATTRIBUTES[] = {
		"cmf/avg_control_unit_queuing_time",
		"cmf/avg_device_active_only_time",
		"cmf/avg_device_busy_time",
		"cmf/avg_device_connect_time",
		"cmf/avg_device_disconnect_time",
		"cmf/avg_function_pending_time",
		"cmf/avg_initial_command_response_time",
		"cmf/avg_sample_interval",
		"cmf/avg_utilization",
		"cmf/ssch_rsch_count",
		"cmf/sample_count",
		NULL
	};
	char buff[256];

	strLength = readSysfsAttr(deviceName, CMB_ENABLE, buff, 256);
	if (strLength < 1) {
		return NULL;
	}
	if (buff[0] != '1') {
		DBGONLY(printf("cmb_enable!=1\n");)
		return NULL;
	}

	for( attributeCount = 0;
	     ATTRIBUTES[attributeCount] != NULL;
	     ++attributeCount);

	char** attributeValues = calloc(attributeCount, 256);
	if (attributeValues == NULL) {
		fprintf(stderr, "calloc() failed\n");
		return NULL;
	}

	int sampleCount1, sampleCount2;
	int loops = 0, len;

	do {
		len = readSysfsAttr(deviceName, SAMPLE_COUNT, buff, 256);
		if (len < 0) {
			free(attributeValues);
			return NULL;
		}

		sampleCount1 = atoi(buff);

		if (sampleCount1 < 0) {
			DBGONLY(printf("sample_count < 0\n");)
			free(attributeValues);
			return NULL;
		}

		strLength = 0;
		for(i = 0; i < attributeCount; ++i) {
			DBGONLY(printf("%s=", ATTRIBUTES[i]);)
			len = readSysfsAttr(deviceName, ATTRIBUTES[i],
					    attributeValues[i], 256);
			if (len < 1) {
				printf("NULL\n");
				strLength+=2;
				continue;
			}
			strLength += max(strlen(attributeValues[i]), 2);
			DBGONLY(printf("%s\n", attributeValues[i]);)
		}

		len = readSysfsAttr(deviceName, SAMPLE_COUNT, buff, 256);
		if (len < 0) {
			free(attributeValues);
			return NULL;
		}
		sampleCount2 = atoi(buff);
		++loops;
		DBGONLY(printf("sample count = %d/%d, loops = %d\n",
			       sampleCount1, sampleCount2, loops);)
	} while (sampleCount2 - sampleCount1 && loops < 2);

	char* result = calloc(strLength+1, sizeof(char));

	if (result==NULL) {
		free(attributeValues);
		return NULL;
	}

	size_t remainder = strLength;

	for(i = 0; i < attributeCount; ++i) {
		char *value = attributeValues[i];

		if (value == NULL || strlen(value)==0) { value = "0\n"; }
		strncat(result, value, remainder);
		remainder -= strlen(value);
		result[strlen(result)-1] = ':';
	}

	free(attributeValues);

	return result;
}
	
int metricRetrECKD( int mid, 
			   MetricReturner mret ) { 
	char dirpath[256];
	int deviceCount = 0;

	DIR *dirfd;
	struct dirent *dent;

	sprintf(dirpath, "/sys/bus/%s/drivers/%s", BUS, DRIVER);
	dirfd = opendir(dirpath);
	if (!dirfd)
		return 0;
	while ((dent = readdir(dirfd)) != NULL) {
		char *values;
		int css, lcss, ccw;

		if (sscanf(dent->d_name,
			   "%d.%d.%04d", &css, &lcss, &ccw) != 3)
			continue;

		values = readSysFs(dent->d_name);
		DBGONLY(printf("device name = %s\n", dent->d_name);)

		if (values!=NULL) {
			DBGONLY(printf("%s\n", values);)
			int result = sendMetric(mid, mret,
						dent->d_name, values);
			free(values);
			if (!result) {
				deviceCount = -1;
				break;
 			}
			++deviceCount;
 		}
 	}
	closedir(dirfd);

	return deviceCount;
}
/* ---------------------------------------------------------------------------*/
/*                          end of metricProcessor.c                          */
/* ---------------------------------------------------------------------------*/

